module shot_extroplate_mod
use shot_kernal_mod
use velocity_mod
use data_mod
use image_mod
use mpi_sep
implicit none

type e_par_type
  integer :: ns,nw,nref
  integer :: nz,nbdx,nbdy
  integer  :: op
  integer  :: MPI_size 

  real:: dx,dy,dw,w0
  real:: dz,fd_a,fd_b,fd_trick
  logical  :: is2d,plane
end type e_par_type

type e_1shot_par_type
  real :: mig_x_min,mig_y_min
  integer :: nx,ny
  integer :: MPI_size
end type e_1shot_par_type


type(e_par_type)::e_par
type(e_1shot_par_type) :: e_1s_par
integer :: n_node,i_node,n_node_big,i_node_nw
contains


subroutine shot_e_init()
integer ierr,size_real,size_logical,size_int
call MPI_INIT(ierr)
call MPI_COMM_RANK( MPI_COMM_WORLD,i_node,ierr)
call MPI_COMM_SIZE( MPI_COMM_WORLD,n_node,ierr)

call MPI_TYPE_SIZE(MPI_REAL,size_real,ierr)
call MPI_TYPE_SIZE(MPI_LOGICAL,size_logical,ierr)
call MPI_TYPE_SIZE(MPI_INTEGER,size_int,ierr)
e_par%MPI_size=size_real*8+size_logical*2+size_int*8
e_1s_par%MPI_size=size_real*2+size_int*3

call data_mod_init(e_par%ns,e_par%nw,e_par%dx,e_par%dy,e_par%dw,e_par%w0,i_node)
!e_par%nw=500
if (i_node==0) then
  call from_param('operator',e_par%op,3)
  call from_param('mig_dz',e_par%dz,6.0)
  call from_param('fd_a',e_par%fd_a,0.5)
  call from_param('fd_b',e_par%fd_b,0.0)
  call from_param('fd_trick',e_par%fd_trick,1.0/6.0)
  call from_param('mig_nz',e_par%nz,800)
  call from_param('nbdx', e_par%nbdx,10)
  call from_param('nbdy',e_par%nbdy,10)
  call from_param('nref_vel',e_par%nref,3)
  call from_param('plane', e_par%plane,.false.)
end if
call velocity_mod_init(e_par%dx,e_par%dy,e_par%dz,e_par%is2d,i_node)
call image_mod_init(e_par%dx,e_par%dy,e_par%dz,e_par%nz,i_node)


call MPI_BCAST(e_par,e_par%MPI_Size,MPI_BYTE,0,MPI_COMM_WORLD,ierr)
!write(0,*)e_par%ns,e_par%nw,e_par%dx,e_par%dy,e_par%dw,e_par%w0
!write(*,*) "is2d", e_par%is2d
i_node_nw=e_par%nw/n_node
n_node_big=mod(e_par%nw,n_node)
!write(*,*) "inode",i_node,"dz",e_par%dz,e_par%nw,e_par%ns,n_node,i_node_nw,n_node_big
end subroutine shot_e_init

subroutine shot_e_run
complex,allocatable :: wld_s(:,:),wld_r(:,:),z0_s(:,:),z0_r(:,:),wavelet(:),tmp_z0_s(:,:)
complex,allocatable :: z0_s_rotate(:,:,:), z0_r_rotate(:,:,:)
complex,allocatable :: z0_r_i_node_nw(:,:,:), tmp_wld_s(:,:,:)
real, allocatable :: vel(:,:,:)
real, allocatable :: image(:,:,:,:,:),img_rev(:,:,:,:),img(:,:,:,:),image_rotate(:,:,:,:,:),image_hz(:,:,:,:,:)
integer :: i_s,i_w,i_z,i_x,i_y
real    :: v0_z(5),w,w0
real    :: dx,dy,dw
integer :: ns,nw,nz,nbdx,nbdy
real    :: dz,fd_a,fd_b,fd_trick
integer :: nx,ny,nx_rotate,ny_rotate,nz_rotate,nz_s_rotate
integer :: nhx,nhy
integer :: op
real :: mig_x_min,mig_y_min,mig_z_min
real :: rotate_angle,pi,rotate_center
integer :: ix_rotate_center
logical :: is2d,no_rotate,plane
type(sep3d)::testimage
type(sep3d)::tmp_vel,wavefld
integer,external :: srite,sseek_block
integer :: ierr,iw_f,iw_e,i_nd,tmp_i_node,wld_block_size
integer :: tmp_i_w,tmp_iw_f,tmp_iw_e,iref,tmp_ivx,tmp_ix,tmp_iy
real :: tmp_wvx, velread_mig_x_min,plane_center
integer :: ix,iy,ii_s
integer :: status(MPI_STATUS_SIZE)
logical :: bb
! read from parameter
bb=.true.
pi=3.1415926 
dz=e_par%dz; fd_a=e_par%fd_a; fd_b=e_par%fd_b; fd_trick=e_par%fd_trick; nz=e_par%nz; nbdx=e_par%nbdx; nbdy=e_par%nbdy; is2d=e_par%is2d
no_rotate=.false.; plane=e_par%plane
if (.not. plane) then
  rotate_angle=0.0/180.0*pi
else
  rotate_angle=45.0/180.0*pi
end if
 rotate_angle=-30.0/180.0*pi
rotate_angle=0.0

! read from data file
dx=e_par%dx; dy=e_par%dy; dw=e_par%dw;ns=e_par%ns;nw=e_par%nw ; w0=e_par%w0; op=e_par%op

allocate(wavelet(nw))
call image_get_nhxnhy(nhx,nhy)
!write(*,*) "image offset number nhx:",nhx, "nhy",nhy

!do i_s=1,ns
!do i_s=1,ns
do i_s=1,ns
  if (i_node==0)then
  write(*,*) "The", i_s, "th shot of", ns, "shots","inode",i_node
  end if
   !get min_x min_y nx ny of data at i_s shot
  if ( .not. plane) then
    call data_1shot_init(i_s,mig_x_min,mig_y_min,nx,ny)
  else
    call plane_1shot_init(i_s,mig_x_min,mig_y_min,nx,ny,plane_center,rotate_angle,i_node)
    !rotate_angle=0.0
  end if
  
  !rotate_angle=0.0/180.0*pi
  
  write(*,*) "rotate_angle=",rotate_angle/pi*180.0

  if (rotate_angle <0)then
    velread_mig_x_min=mig_x_min+(nx-1)*dx ! when rotate angle <0, velread_mig_x_min acturally is mig_x_max
  else
    velread_mig_x_min=mig_x_min  
  end if
   !write(*,*) "min mig_x of this shot x:",mig_x_min, "y:",mig_y_min 
  ! mig_x_min=mig_x_min-abs(dx)*nbdx
  !if (.not.e_par%is2d) mig_y_min=mig_y_min-abs(dy)*nbdy
  e_1s_par%mig_x_min=mig_x_min;  e_1s_par%mig_y_min=mig_y_min
  !nx=nx+2*nbdx
  !if (.not. e_par%is2d) ny=ny+2*nbdy
  if (no_rotate) then
     nx_rotate=nx; nz_rotate=nz; if (.not. e_par%is2d) ny_rotate=ny; nz_s_rotate=nz
  else
     nz_rotate=((nz-1)*dz*cos(abs(rotate_angle))+(nx-1)*dx*sin(abs(rotate_angle))-0)/dz+1  
     ix_rotate_center=((nz-1)*dz*sin(abs(rotate_angle))-0.0)/dx
     if (nbdx> ix_rotate_center) ix_rotate_center=nbdx
     nx_rotate=((nx-1)*dx*cos(abs(rotate_angle))-0)/dx+1+ix_rotate_center+nbdx
     ny_rotate=1
     if (mod(nx_rotate,16)/=0) nx_rotate=nx_rotate/16*16 
     if (.not. e_par%is2d) ny_rotate=ny 
     nz_s_rotate=((nx-1)*dx*sin(abs(rotate_angle))-0)/dz+1
  end if
  write(*,*) "nx",nx_rotate
  ny_rotate=1
  e_1s_par%nx=nx;  e_1s_par%ny=ny  !wait
  !write(*,*) "nx,ny", e_1s_par%nx,"ny",e_1s_par%ny
  !write(*,*)"nz",nz_rotate,"nzs",nz_s_rotate, "nx",nx_rotate, "ny",ny_rotate
  mig_z_min=0.0
  allocate(z0_s(nx,ny),z0_r(nx,ny),z0_r_i_node_nw(nx,ny,i_node_nw+1))
  allocate(tmp_z0_s(nx,ny))
  ! wait to get nx,ny after rotate
  call shot_ker_init(nx_rotate,ny_rotate,e_par%nref) 
  call ker_p_init(op,nx_rotate,ny_rotate,dx,dy,dz,fd_a,fd_b,fd_trick,is2d)
  ! write(*,*)"nz,nx,ny",nz,nx,ny
  allocate(wld_s(nx_rotate,ny_rotate),wld_r(nx_rotate,ny_rotate))
  allocate(tmp_wld(nx_rotate,ny_rotate),add_wld(nx_rotate,ny_rotate))
  allocate(vel(nz_rotate,nx_rotate,ny_rotate),image(nhx,nhy,nx_rotate,ny_rotate,nz_rotate),img(nhx,nhy,nx_rotate,ny_rotate) )
  allocate(image_rotate(nz,nhx,nhy,nx+nbdx,ny),image_hz(nz,nhx,nhy,nx+nbdx,ny))
  allocate(z0_s_rotate(nx_rotate,ny_rotate,nz_s_rotate),z0_r_rotate(nx_rotate,ny_rotate,nz_s_rotate))
  allocate(tmp_wld_s(nz_rotate,nx_rotate,ny_rotate))
  if (i_node==0) allocate(img_rev(nhx,nhy,nx_rotate,ny_rotate))
  if (i_node==0) then
    call velocity_read(velread_mig_x_min,mig_y_min,mig_z_min,nx_rotate,ny_rotate,nz_rotate,vel,e_par%is2d,rotate_angle,ix_rotate_center)
  
  end if
  if (i_node==0 .and. i_s==1) then
   call init_sep3d(tmp_vel, "OUTPUT","FLOAT", "REGULAR", ctag="Tmpvel"); call axis_allocate(tmp_vel,3)
      tmp_vel%o=0; tmp_vel%d=1  !dz;        tmp_vel%d(2)=dx;        tmp_vel%d(3)=dy
      tmp_vel%n(1)=nz_rotate; tmp_vel%n(2)=nx_rotate; tmp_vel%n(3)=ny_rotate
      tmp_vel%d(2)=dx; tmp_vel%d(1)=dz; tmp_vel%o(2)=mig_x_min-ix_rotate_center*dx
   call sep3d_set_sep3d(tmp_vel); call sep3d_write_description("Tmpvel",tmp_vel)
   if (srite("Tmpvel", vel,nz_rotate*nx_rotate*ny_rotate*4)/=nz_rotate*nx_rotate*ny_rotate*4) call seperr("write tmp vel w")
  end if
  if (bb) then
  wld_block_size=8*nx_rotate*nz_rotate
  
  call MPI_BCAST(vel,nz_rotate*nx_rotate*ny_rotate,MPI_REAL,0,MPI_COMM_WORLD,ierr)
!  write(*,*) "vel",i_node,"max",maxval(vel),"min",minval(vel)
  if (i_node==0) then
    do i_nd=1,n_node-1
      call read_data_for_i_node(i_nd,mig_x_min,mig_y_min,nx,ny,i_s,z0_r,z0_r_i_node_nw)
      call MPI_Send(z0_r_i_node_nw,nx*ny*(i_node_nw+1),MPI_COMPLEX,i_nd,i_s,MPI_COMM_WORLD,ierr)
    end do
    i_nd=0
    call read_data_for_i_node(i_nd,mig_x_min,mig_y_min,nx,ny,i_s,z0_r,z0_r_i_node_nw)
  else
    call MPI_Recv(z0_r_i_node_nw,nx*ny*(i_node_nw+1),MPI_COMPLEX,0,i_s,MPI_COMM_WORLD,status,ierr)
  end if

  call get_fe_w(i_node,iw_f,iw_e)
  image=0.0
  write(*,*) "finish init"
  do i_w=iw_f,iw_e
    w=(i_w-1)*dw+w0
    tmp_wld_s=0.0 
    write(*,*)'i_w',i_w,'i_node',i_node,"shots",i_s
    z0_s=0.0
!!!    ! get source and receiver data at the surface: z0_s,z0_r
    if ( .not. plane) then
      call data_source(mig_x_min,mig_y_min,nx,ny,i_w,i_s,z0_s)
    else
      !plane_center=5000.0 for marmousi green
      !plane_center=62800.0
      call plane_source(mig_x_min,mig_y_min,nx,ny,i_w,i_s,z0_s,plane_center)
      !tmp_z0_s=0.0
      !do ii_s=1,ns
      !  write(*,*) "ii_s",ii_s
      !  call plane_source(mig_x_min,mig_y_min,nx,ny,i_w,ii_s,z0_s,plane_center)
      !  tmp_z0_s=tmp_z0_s+z0_s
      !end do
      !z0_s=tmp_z0_s
    end if

    z0_r=z0_r_i_node_nw(:,:,i_w-iw_f+1)
    if (no_rotate) then
       wld_s=z0_s; wld_r=z0_r
    else
      call rotate(z0_s,z0_s_rotate,rotate_angle,ix_rotate_center,dx,dz,nx,nx_rotate,nz_s_rotate)
      call rotate(z0_r,z0_r_rotate,rotate_angle,ix_rotate_center,dx,dz,nx,nx_rotate,nz_s_rotate)
    end if
    wld_s=0.0; wld_r=0.0 
    do i_z=1,nz_rotate
      if (i_z <=nz_s_rotate) then
        wld_s=wld_s+z0_s_rotate(:,:,i_z)
        wld_r=wld_r+z0_r_rotate(:,:,i_z)
      end if
      call shot_ker_depth_onestep(wld_s,vel(i_z,:,:),w,-1)
      call shot_ker_depth_onestep(wld_r,vel(i_z,:,:),w,+1)
!      wld_r=cmplx(cos(w*1.0),sin(w*1.0)) ! for marmousi green
!      wld_r=cmplx(cos(w*3.0),sin(w*3.0))  ! for zigg green
      call image_cross(wld_s,wld_r,image(:,:,:,:,i_z),nx_rotate,ny_rotate)  
      tmp_wld_s(i_z,:,:)=wld_s
    end do 
  end do
  

  do i_z=1,nz_rotate
    img=image(:,:,:,:,i_z)
    call MPI_Reduce(img,img_rev,nhx*nhy*nx_rotate*ny_rotate,MPI_REAL,MPI_SUM,0,MPI_COMM_WORLD,ierr)
    if (i_node==0) image(:,:,:,:,i_z)=img_rev
  end do
  
  

  !if (i_node==0 .and. i_s==1) then
  !  call init_sep3d(testimage, "OUTPUT","FLOAT", "REGULAR", ctag="Testimage"); call axis_allocate(testimage,2)
  !  testimage%n(1)=nx_rotate; testimage%n(2)=nz_rotate
  !  testimage%d(1)=dx; testimage%d(2)=dz; testimage%o(1)=mig_x_min-ix_rotate_center*dx; testimage%o(2)=0.0
  !  call sep3d_set_sep3d(testimage); call sep3d_write_description("Testimage",testimage)
  !  if (srite("Testimage", image,nz_rotate*nx_rotate*4)/=nz_rotate*nx_rotate*4) call seperr("write tmp vel w")
  !end if



  if (i_node==0) call img_rotate(image,image_rotate,rotate_angle,ix_rotate_center,dx,dz,nx+nbdx,nz,nx_rotate,nz_rotate)

  !if (i_node==0 ) call image_stack(image(:,:,:,:,:),mig_x_min-ix_rotate_center*dx,mig_y_min,nx_rotate,ny)

!for both h and v offset   
  if (i_node==0) then
    write(*,*) "begin decompose"
    call image_decompose(image_rotate,image_hz,nx+nbdx,nz,rotate_angle)
    write(*,*) "end decompose"
  end if
  if (i_node==0) then
    write(*,*) "begin write"
    if (rotate_angle >=0 ) then
      call image_stack(image_rotate,mig_x_min,mig_y_min,nx+nbdx,ny)
      ! for v offset
      call image_stack_hz(image_hz,mig_x_min,mig_y_min,nx+nbdx,ny)
    else
      call image_stack(image_rotate,mig_x_min-nbdx*dx,mig_y_min,nx+nbdx,ny)
      ! for v offset
      call image_stack_hz(image_hz,mig_x_min-nbdx*dx,mig_y_min,nx+nbdx,ny)
    end if
    write(*,*) "end write"
  end if
! for both h and v offset 
  call shot_ker_release 
  end if
  deallocate(wld_s,wld_r,z0_s,z0_r,z0_r_i_node_nw,vel,image,img,z0_s_rotate,z0_r_rotate,tmp_wld_s,image_rotate,image_hz)
  deallocate(tmp_z0_s)
  if (i_node==0) deallocate(img_rev)
end do
deallocate(wavelet)
end subroutine shot_e_run

! calculate first and end frequency of every node
subroutine get_fe_w(i_nd,iw_f,iw_e)
integer ::i_nd,iw_f,iw_e
  if (i_nd /= 0) then
    if (i_nd <= n_node_big) then
      iw_f=(i_nd-1)*(i_node_nw+1)+1
      iw_e=i_nd*(i_node_nw+1)
    else
      iw_f=n_node_big*(i_node_nw+1)+(i_nd-n_node_big-1)*i_node_nw+1
      iw_e=n_node_big*(i_node_nw+1)+(i_nd-n_node_big)*i_node_nw
    end if
  else
    iw_f=n_node_big*(i_node_nw+1)+(n_node-n_node_big-1)*i_node_nw+1
    iw_e=n_node_big*(i_node_nw+1)+(n_node-n_node_big)*i_node_nw
  end if
end subroutine get_fe_w
! calculate first and end frequency of every node

! read data for every node which will be sent out by MPI_Send
subroutine read_data_for_i_node(i_nd,mig_x_min,mig_y_min,nx,ny,i_s,z0_r,z0_r_i_node_nw)
integer :: i_nd
real    :: mig_x_min,mig_y_min
integer :: nx,ny,i_s
complex :: z0_r(:,:),z0_r_i_node_nw(:,:,:)

integer :: iw_f,iw_e,i_w

z0_r_i_node_nw=0.0
call get_fe_w(i_nd,iw_f,iw_e)
do i_w=iw_f,iw_e
  if ( .not. e_par%plane) then 
    call  data_receiver(mig_x_min,mig_y_min,nx,ny,i_w,i_s,z0_r)
  else
    call plane_receiver(mig_x_min,mig_y_min,nx,ny,i_w,i_s,z0_r)
  end if
  z0_r_i_node_nw(:,:,i_w-iw_f+1)=z0_r(:,:)
end do
end subroutine read_data_for_i_node
! read data for every node which will be sent out by MPI_Send



real function average(vvel,nx,ny)
real :: vvel(:,:)
integer :: nx,ny

real :: v0_z
integer :: i_x,i_y
v0_z=0.0
do i_y=1,ny
  do i_x=1,nx
    v0_z=v0_z+vvel(i_x,i_y)
  end do
end do
average=v0_z/(nx*ny)
return
end function average

subroutine shot_e_release
integer :: ierr
!call velocity_mod_release
call MPI_FINALIZE(ierr)
end subroutine shot_e_release

end module shot_extroplate_mod
