module image_mod
use sep3d_struct_mod
use sep
use mpi_sep
use sinc_f90_mod
implicit none

type image_par_type
  real :: min_x,max_x,min_y,max_y
  real :: dx,dy,dz
  integer :: nx,ny,nz,nhx,nhy
  integer :: MPI_size
end type image_par_type

type(sep3d),private :: image,image_hz
type(image_par_type) :: image_par

real :: sincx_table(8,101)
contains

subroutine image_mod_init(dx,dy,dz,nz,i_node)
integer,external :: auxin
real :: dx,dy,dz
integer ::nz,i_node
real,allocatable :: timg(:)
integer :: ix,iy,ihx,ihy
integer,external :: srite
integer :: ierr,size_real,size_int

do ihx=1,101
 call sinc_f90_mksinc((ihx-1)*0.01,8,sincx_table(:,ihx)) 
end do

call MPI_TYPE_SIZE(MPI_REAL,size_real,ierr)
call MPI_TYPE_SIZE(MPI_INTEGER,size_int,ierr)
image_par%MPI_size=7*size_real+6*size_int
if (i_node==0) then
  allocate(timg(nz))
  timg=0.0
  call from_param('image_min_x',image_par%min_x,1000.0)
  call from_param('image_max_x',image_par%max_x,10000.0)
  call from_param('image_min_y',image_par%min_y,0.0)
  call from_param('image_max_y',image_par%max_y,0.0)
  call from_param('nhx',image_par%nhx, 1)
  call from_param('nhy',image_par%nhy, 1)
  image_par%dx=abs(dx)
  image_par%dy=abs(dy)
  image_par%dz=dz
  image_par%nx=(image_par%max_x-image_par%min_x)/(image_par%dx)+1
  image_par%ny=(image_par%max_y-image_par%min_y)/(image_par%dy)+1
  image_par%nz=nz

  call init_sep3d(image, "OUTPUT","FLOAT", "REGULAR", ctag="Image")
  call axis_allocate(image,5)
  ! image:1:z 2:x, 3:y, 4:hx, 5:hy
  !z:
  image%o(1)=0.0; image%d(1)=image_par%dz; image%n(1)=image_par%nz
  !hx:
  !wait for dx dy
  image%o(2)=-image_par%dx*image_par%nhx/2; image%d(2)=image_par%dx; image%n(2)=image_par%nhx
  !hy:
  image%o(3)=-image_par%dy*image_par%nhy/2; image%d(3)=image_par%dy; image%n(3)=image_par%nhy
  !x:
  image%o(4)=image_par%min_x;image%d(4)=(image_par%dx);image%n(4)=image_par%nx
  !y:
  image%o(5)=image_par%min_y;image%d(5)=(image_par%dy);image%n(5)=image_par%ny

  call sep3d_set_sep3d(image)
  call sep3d_write_description("Image",image)


  call init_sep3d(image_hz, "OUTPUT","FLOAT", "REGULAR", ctag="Image_hz")
  call axis_allocate(image_hz,5)
  ! image:1:z 2:x, 3:y, 4:hx, 5:hy
  !z:
  image_hz%o(1)=0.0; image_hz%d(1)=image_par%dz; image_hz%n(1)=image_par%nz
  !hx:
  !wait for dx dy
  image_hz%o(2)=-image_par%dx*image_par%nhx/2; image_hz%d(2)=image_par%dx; image_hz%n(2)=image_par%nhx
  !hy:
  image_hz%o(3)=-image_par%dy*image_par%nhy/2; image_hz%d(3)=image_par%dy; image_hz%n(3)=image_par%nhy
  !x:
  image_hz%o(4)=image_par%min_x;image_hz%d(4)=(image_par%dx);image_hz%n(4)=image_par%nx
  !y:
  image_hz%o(5)=image_par%min_y;image_hz%d(5)=(image_par%dy);image_hz%n(5)=image_par%ny

  call sep3d_set_sep3d(image_hz)
  call sep3d_write_description("Image_hz",image_hz)



  do iy=1,image_par%ny
    do ix=1,image_par%nx
      do ihy=1,image_par%nhy
        do ihx=1,image_par%nhx
           if ( srite("Image",timg,image_par%nz*4)/=image_par%nz*4 ) call seperr("err init image")
        end do
      end do
    end do
  end do

  do iy=1,image_par%ny
    do ix=1,image_par%nx
      do ihy=1,image_par%nhy
        do ihx=1,image_par%nhx
           if ( srite("Image_hz",timg,image_par%nz*4)/=image_par%nz*4 ) call seperr("err init image")
        end do
      end do
    end do
  end do

  deallocate(timg)

  call sep3df_delete(image,"Image")
  call init_sep3d("Image",image,"SCRATCH")
  call sep3df_delete(image_hz,"Image_hz")
  call init_sep3d("Image_hz",image_hz,"SCRATCH")



end if
call MPI_BCAST(image_par,image_par%MPI_Size,MPI_BYTE,0,MPI_COMM_WORLD,ierr)
!write(*,*)"image",image_par%min_x
end subroutine image_mod_init


subroutine image_get_nhxnhy(nhx,nhy)
integer nhx,nhy
nhx=image_par%nhx
nhy=image_par%nhy
end subroutine image_get_nhxnhy



subroutine image_cross(wld_s,wld_r,image_z,nx,ny)
complex :: wld_s(:,:),wld_r(:,:)
real    :: image_z(:,:,:,:)
integer :: nx,ny
complex :: ws,wr
integer :: ihx,ihy,ix,iy
integer :: shift_ix_s,shift_iy_s,shift_ix_r,shift_iy_r
do ihy=1,image_par%nhy
  do ihx=1,image_par%nhx
    do iy=1,ny
      do ix=1,nx
        shift_ix_s=ix-(-((image_par%nhx-1)/2+1)+ihx)
        shift_iy_s=iy-(-((image_par%nhy-1)/2+1)+ihy)
        shift_ix_r=ix+(-((image_par%nhx-1)/2+1)+ihx)
        shift_iy_r=iy+(-((image_par%nhy-1)/2+1)+ihy)
        if (shift_ix_s <1) shift_ix_s=1; if (shift_ix_s >nx) shift_ix_s=nx
        if (shift_ix_r <1) shift_ix_r=1; if (shift_ix_r >nx) shift_ix_r=nx
        if (shift_iy_s <1) shift_iy_s=1; if (shift_iy_s >ny) shift_iy_s=ny
        if (shift_iy_r <1) shift_iy_r=1; if (shift_iy_r >ny) shift_iy_r=ny
        ws=wld_s(shift_ix_s,shift_iy_s)
        wr=wld_r(shift_ix_r,shift_iy_r)  
        !ws=wld_s(ix-(-((image_par%nhx-1)/2+1)+ihx),iy-(-((image_par%nhy-1)/2+1)+ihy))
        !wr=wld_r(ix+(-((image_par%nhx-1)/2+1)+ihx),iy+(-((image_par%nhy-1)/2+1)+ihy))
        image_z(ihx,ihy,ix,iy)=image_z(ihx,ihy,ix,iy)+real( ws*conjg(wr))
      end do
    end do
  end do  
end do
end subroutine image_cross



subroutine image_decompose(image_hx,image_hz,nx,nz,rotate_angle)
real :: image_hx(:,:,:,:,:),image_hz(:,:,:,:,:)
integer :: nx,nz
real    :: rotate_angle

real,allocatable :: image_dip(:) 

integer :: ix,iz,i_ihz,i_ihx,ihdip,ihx,ihz,ii_hz,ii_hx,iwx
real :: hx,hz,whx,whz,ohx,hdip

real :: ohx,ddx,wwx
real :: sincz(8),sincx(8)
integer :: nit

nit=8

allocate( image_dip(image_par%nhx*nit) )

ohx=-image_par%dx*image_par%nhx/2

ddx=image_par%dx/nit

write(*,*) "dddddd4"
image_hz=0.0
do ix=1,nx
  do iz=1,nz
    image_dip=0.0
    do ihdip=1,image_par%nhx*nit
      hdip=(ihdip-1)*ddx+ohx
      ihx=(hdip-ohx)/image_par%dx+1
      whx=(hdip-(ohx+(ihx-1)*image_par%dx))/image_par%dx
      iwx=(whx-0.0)/0.01+1
      wwx=(whx- ((iwx-1)*0.01))/0.01
      sincx=(1-wwx)*sincx_table(:,iwx)+wwx*sincx_table(:,iwx+1)
      do i_ihx=1,8
        ii_hx=i_ihx+ihx-4
        if (ii_hx>0 .and. ii_hx <= image_par%nhx ) then
         image_dip(ihdip)=image_dip(ihdip)+sincx(i_ihx)*image_hx(iz,ii_hx,1,ix,1)
        end if
      end do
    end do

    image_hx(iz,:,1,ix,1)=0.0
    image_hz(iz,:,1,ix,1)=0.0
    do ihdip=1,image_par%nhx*nit
      hdip=ohx+(ihdip-1)*ddx
      if (hdip==0) then
        hx=0.0; hz=0.0
      else
        hx=hdip/cos(rotate_angle)
        hz=hdip/sin(rotate_angle)
      end if
      ihx=(hx-ohx)/image_par%dx+1
      ihz=(hz-ohx)/image_par%dx+1
    
      if (ihx > 0 .and. ihx<=image_par%nhx ) then 
        whx=( hx-(ohx+(ihx-1)*image_par%dx) )/image_par%dx
        iwx=(whx-0.0)/0.01+1
        wwx=(whx- ((iwx-1)*0.01))/0.01
        sincx(:)=(1-wwx)*sincx_table(:,iwx)+wwx*sincx_table(:,iwx+1)
        do i_ihx=1,8
          ii_hx=i_ihx+ihx-4
          if (ii_hx >0 .and. ii_hx <= image_par%nhx ) then
            image_hx(iz,ii_hx,1,ix,1)= image_hx(iz,ii_hx,1,ix,1)+sincx(i_ihx)*image_dip(ihdip)
          end if
        end do 
      end if
     
      if (ihz >0 .and. ihz <=image_par%nhx ) then
        whz=( hz-(ohx+(ihz-1)*image_par%dx) )/image_par%dx
        iwx=(whz-0.0)/0.01+1
        wwx=(whz-((iwx-1)*0.01))/0.01
        sincz=(1-wwx)*sincx_table(:,iwx)+wwx*sincx_table(:,iwx+1)
        do i_ihz=1,8
          ii_hz=i_ihz+ihz-4
          if (ii_hz >0 .and. ii_hz <= image_par%nhx ) then
            image_hz(iz,ii_hz,1,ix,1)=image_hz(iz,ii_hz,1,ix,1)+ sincz(i_ihz)*image_dip(ihdip)
          end if
        end do
      end if

    end do
  end do
end do
write(*,*)"nddddddddd4"
deallocate (image_dip)
end subroutine image_decompose


subroutine image_stack(img,mig_x_min,mig_y_min,nx,ny)
real :: img(:,:,:,:,:)
real :: mig_x_min,mig_y_min
integer :: nx,ny
real,allocatable:: tmp_img(:,:,:,:)

integer :: ix,iy,img_ix,img_iy,nx_rw
integer :: tmp_img_ix, tmp_nx,tmp_img_bix
integer :: block_size, block_i
integer :: sseek_block,sreed,srite
img_ix=(mig_x_min-image_par%min_x)/(image_par%dx)+1
if (img_ix <1 ) then
  tmp_img_ix=1; tmp_nx=nx-1+img_ix; tmp_img_bix=1-img_ix+1
else
  tmp_img_ix=img_ix; tmp_nx=nx; tmp_img_bix=1
end if

if (tmp_img_ix+tmp_nx-1>image_par%nx) then
   nx_rw=image_par%nx-tmp_img_ix+1
else
   nx_rw=nx
end if

if (.not. nx_rw <1) then
  allocate(tmp_img(image_par%nz,image_par%nhx,image_par%nhy,nx_rw))
  block_size=4*image_par%nz*image_par%nhx*image_par%nhy
  do iy=1,ny
    tmp_img=0.0
    img_iy=((iy-1)*image_par%dy+mig_y_min-image_par%min_y)/image_par%dy+1
    if (img_iy <= image_par%ny .and. img_iy> 0) then
      block_i=(img_iy-1)*image_par%nx+tmp_img_ix-1
      !write(*,*) "nynynynyny",iy,ny,block_size,block_i,img_iy,tmp_img_ix,nx
      if ( sseek_block("Image",block_i,block_size,0)/=block_i) call seperr("Image seek wrong")
      if (sreed("Image",tmp_img,block_size*nx_rw)/=block_size*nx_rw )call seperr("Image read wrong ?")
      ! wait get rid of image_par%nz
      tmp_img=tmp_img+img(1:image_par%nz,:,:,tmp_img_bix:tmp_img_bix+nx_rw-1,iy)
      if ( sseek_block("Image",block_i,block_size,0)/=block_i) call seperr("Image seek wrong")
      if (srite("Image",tmp_img,block_size*nx_rw)/=block_size*nx_rw) call seperr("Image write wrong")
    end if
  end do
  deallocate(tmp_img)
end if
end subroutine image_stack


subroutine image_stack_hz(img,mig_x_min,mig_y_min,nx,ny)
real :: img(:,:,:,:,:)
real :: mig_x_min,mig_y_min
integer :: nx,ny
real,allocatable:: tmp_img(:,:,:,:)

integer :: ix,iy,img_ix,img_iy,nx_rw
integer :: tmp_img_ix, tmp_nx,tmp_img_bix
integer :: block_size, block_i
integer :: sseek_block,sreed,srite
img_ix=(mig_x_min-image_par%min_x)/(image_par%dx)+1
if (img_ix <1 ) then
  tmp_img_ix=1; tmp_nx=nx-1+img_ix; tmp_img_bix=1-img_ix+1
else
  tmp_img_ix=img_ix; tmp_nx=nx; tmp_img_bix=1
end if

if (tmp_img_ix+tmp_nx-1>image_par%nx) then
   nx_rw=image_par%nx-tmp_img_ix+1
else
   nx_rw=nx
end if

if (.not. nx_rw <1) then
  allocate(tmp_img(image_par%nz,image_par%nhx,image_par%nhy,nx_rw))
  block_size=4*image_par%nz*image_par%nhx*image_par%nhy
  do iy=1,ny
    tmp_img=0.0
    img_iy=((iy-1)*image_par%dy+mig_y_min-image_par%min_y)/image_par%dy+1
    if (img_iy <= image_par%ny .and. img_iy> 0) then
      block_i=(img_iy-1)*image_par%nx+tmp_img_ix-1
      !write(*,*) "nynynynyny",iy,ny,block_size,block_i,img_iy,tmp_img_ix,nx
      if ( sseek_block("Image_hz",block_i,block_size,0)/=block_i) call seperr("Image seek wrong")
      if (sreed("Image_hz",tmp_img,block_size*nx_rw)/=block_size*nx_rw )call seperr("Image read wrong ?")
      ! wait get rid of image_par%nz
      tmp_img=tmp_img+img(1:image_par%nz,:,:,tmp_img_bix:tmp_img_bix+nx_rw-1,iy)
      if ( sseek_block("Image_hz",block_i,block_size,0)/=block_i) call seperr("Image seek wrong")
      if (srite("Image_hz",tmp_img,block_size*nx_rw)/=block_size*nx_rw) call seperr("Image write wrong")
    end if
  end do
  deallocate(tmp_img)
end if
end subroutine image_stack_hz


end module image_mod
